@@ -31,15 +31,14 @@ import sys
 import threading
 import time
 from functools import reduce
-from types import TracebackType
-from typing import Callable  # noqa # used in type hints
-from typing import Dict, List, Optional, Tuple, Type, Union
+from typing import AnyStr, Dict, List, Optional, Union, cast
+from typing import Callable, Set, Tuple  # noqa # used in type hints
 
 import ifaddr
 
 __author__ = 'Paul Scott-Murphy, William McBrine'
 __maintainer__ = 'Jakub Stasiak <jakub@stasiak.at>'
-__version__ = '0.21.0'
+__version__ = '0.22.0'
 __license__ = 'LGPL'
 
 
@@ -63,18 +62,20 @@ if log.level == logging.NOTSET:
 
 # Some timing constants
 
-_UNREGISTER_TIME = 125
-_CHECK_TIME = 175
-_REGISTER_TIME = 225
-_LISTENER_TIME = 200
-_BROWSER_TIME = 500
+_UNREGISTER_TIME = 125  # ms
+_CHECK_TIME = 175  # ms
+_REGISTER_TIME = 225  # ms
+_LISTENER_TIME = 200  # ms
+_BROWSER_TIME = 1000  # ms
+_BROWSER_BACKOFF_LIMIT = 3600  # s
 
 # Some DNS constants
 
 _MDNS_ADDR = '224.0.0.251'
 _MDNS_PORT = 5353
 _DNS_PORT = 53
-_DNS_TTL = 120  # two minutes default TTL as recommended by RFC6762
+_DNS_HOST_TTL = 120  # two minute for host records (A, SRV etc) as-per RFC6762
+_DNS_OTHER_TTL = 4500  # 75 minutes for non-host records (PTR, TXT etc) as-per RFC6762
 
 _MAX_MSG_TYPICAL = 1460  # unused
 _MAX_MSG_ABSOLUTE = 8966
@@ -317,11 +318,9 @@ class BadTypeInNameException(Error):
 
 # implementation classes
 
-OptionalExcInfo = Tuple[Optional[Type[BaseException]], Optional[BaseException], Optional[TracebackType]]
-
 
 class QuietLogger:
-    _seen_logs = {}  # type: Dict[str, OptionalExcInfo]
+    _seen_logs = {}  # type: Dict[str, tuple]
 
     @classmethod
     def log_exception_warning(cls, logger_data=None):
@@ -335,7 +334,7 @@ class QuietLogger:
             logger = log.debug
         if logger_data is not None:
             logger(*logger_data)
-        logger('Exception occurred:', exc_info=exc_info)
+        logger('Exception occurred:', exc_info=True)
 
     @classmethod
     def log_warning_once(cls, *args):
@@ -353,7 +352,7 @@ class DNSEntry:
 
     """A DNS entry"""
 
-    def __init__(self, name, type_, class_):
+    def __init__(self, name: str, type_: int, class_):
         self.key = name.lower()
         self.name = name
         self.type = type_
@@ -381,7 +380,7 @@ class DNSEntry:
         """Type accessor"""
         return _TYPES.get(t, "?(%s)" % t)
 
-    def to_string(self, hdr, other):
+    def to_string(self, hdr, other) -> str:
         """String representation with additional information"""
         result = "%s[%s,%s" % (hdr, self.get_type(self.type),
                                self.get_class_(self.class_))
@@ -419,7 +418,7 @@ class DNSRecord(DNSEntry):
 
     """A DNS record - like a DNS entry, but has a TTL"""
 
-    def __init__(self, name, type_, class_, ttl):
+    def __init__(self, name, type_, class_, ttl: float):
         DNSEntry.__init__(self, name, type_, class_)
         self.ttl = ttl
         self.created = current_time_millis()
@@ -445,7 +444,7 @@ class DNSRecord(DNSEntry):
         and if its TTL is at least half of this record's."""
         return self == other and other.ttl > (self.ttl / 2)
 
-    def get_expiration_time(self, percent):
+    def get_expiration_time(self, percent: float) -> float:
         """Returns the time at which this record will have expired
         by a certain percentage."""
         return self.created + (percent * self.ttl * 10)
@@ -454,7 +453,7 @@ class DNSRecord(DNSEntry):
         """Returns the remaining TTL in seconds."""
         return max(0, (self.get_expiration_time(100) - now) / 1000.0)
 
-    def is_expired(self, now) -> bool:
+    def is_expired(self, now: float) -> bool:
         """Returns true if this record has expired."""
         return self.get_expiration_time(100) <= now
 
@@ -643,10 +642,10 @@ class DNSIncoming(QuietLogger):
         """Constructor from string holding bytes of packet"""
         self.offset = 0
         self.data = data
-        self.questions = []
-        self.answers = []
+        self.questions = []  # type: List[DNSQuestion]
+        self.answers = []  # type: List[DNSRecord]
         self.id = 0
-        self.flags = 0
+        self.flags = 0  # type: int
         self.num_questions = 0
         self.num_answers = 0
         self.num_authorities = 0
@@ -712,7 +711,7 @@ class DNSIncoming(QuietLogger):
             domain = self.read_name()
             type_, class_, ttl, length = self.unpack(b'!HHiH')
 
-            rec = None
+            rec = None  # type: Optional[DNSRecord]
             if type_ == _TYPE_A:
                 rec = DNSAddress(
                     domain, type_, class_, ttl, self.read_string(4))
@@ -799,15 +798,15 @@ class DNSOutgoing:
         self.id = 0
         self.multicast = multicast
         self.flags = flags
-        self.names = {}
-        self.data = []
+        self.names = {}  # type: Dict[str, int]
+        self.data = []  # type: List[bytes]
         self.size = 12
         self.state = self.State.init
 
-        self.questions = []
-        self.answers = []
-        self.authorities = []
-        self.additionals = []
+        self.questions = []  # type: List[DNSQuestion]
+        self.answers = []  # type: List[Tuple[DNSEntry, float]]
+        self.authorities = []  # type: List[DNSPointer]
+        self.additionals = []  # type: List[DNSAddress]
 
     def __repr__(self):
         return '<DNSOutgoing:{%s}>' % ', '.join([
@@ -1050,7 +1049,7 @@ class DNSCache:
     """A cache of DNS entries"""
 
     def __init__(self):
-        self.cache = {}
+        self.cache = {}  # type: Dict[str, List[DNSEntry]]
 
     def add(self, entry):
         """Adds an entry"""
@@ -1124,7 +1123,7 @@ class Engine(threading.Thread):
         threading.Thread.__init__(self, name='zeroconf-Engine')
         self.daemon = True
         self.zc = zc
-        self.readers = {}  # maps socket to reader
+        self.readers = {}  # type: Dict[socket.socket, Listener]
         self.timeout = 5
         self.condition = threading.Condition()
         self.start()
@@ -1231,14 +1230,14 @@ class Reaper(threading.Thread):
 
 class Signal:
     def __init__(self):
-        self._handlers = []
+        self._handlers = []  # type: List[Callable[..., None]]
 
-    def fire(self, **kwargs):
+    def fire(self, **kwargs) -> None:
         for h in list(self._handlers):
             h(**kwargs)
 
     @property
-    def registration_interface(self):
+    def registration_interface(self) -> 'SignalRegistrationInterface':
         return SignalRegistrationInterface(self._handlers)
 
 
@@ -1261,6 +1260,14 @@ class RecordUpdateListener:
         raise NotImplementedError()
 
 
+class ServiceListener:
+    def add_service(self, zc, type_, name) -> None:
+        raise NotImplementedError()
+
+    def remove_service(self, zc, type_, name) -> None:
+        raise NotImplementedError()
+
+
 class ServiceBrowser(RecordUpdateListener, threading.Thread):
 
     """Used to browse for a service of a specific type.
@@ -1273,7 +1280,7 @@ class ServiceBrowser(RecordUpdateListener, threading.Thread):
                  addr: str = _MDNS_ADDR, port: int = _MDNS_PORT, delay: int = _BROWSER_TIME) -> None:
         """Creates a browser for a specific type"""
         assert handlers or listener, 'You need to specify at least one handler'
-        if not type_.endswith(service_type_name(type_)):
+        if not type_.endswith(service_type_name(type_, allow_underscores=True)):
             raise BadTypeInNameException
         threading.Thread.__init__(
             self, name='zeroconf-ServiceBrowser_' + type_)
@@ -1378,22 +1385,23 @@ class ServiceBrowser(RecordUpdateListener, threading.Thread):
 
                 self.zc.send(out, addr=self.addr, port=self.port)
                 self.next_time = now + self.delay
-                self.delay = min(20 * 1000, self.delay * 2)
+                self.delay = min(_BROWSER_BACKOFF_LIMIT * 1000, self.delay * 2)
 
             if len(self._handlers_to_call) > 0 and not self.zc.done:
                 handler = self._handlers_to_call.pop(0)
                 handler(self.zc)
 
 
-ServicePropertiesType = Dict[bytes, Union[bool, str]]
+ServicePropertiesType = Dict[AnyStr, Union[None, bool, AnyStr, float]]
 
 
 class ServiceInfo(RecordUpdateListener):
 
     """Service information"""
 
-    def __init__(self, type_: str, name: str, address: bytes = None, port: int = None, weight: int = 0,
-                 priority: int = 0, properties=None, server: str = None) -> None:
+    def __init__(self, type_: str, name: str, address: Optional[bytes] = None, port: Optional[int] = None,
+                 weight: int = 0, priority: int = 0, properties=b'', server: Optional[str] = None,
+                 host_ttl: int = _DNS_HOST_TTL, other_ttl: int = _DNS_OTHER_TTL) -> None:
         """Create a service description.
 
         type_: fully qualified service type name
@@ -1404,9 +1412,11 @@ class ServiceInfo(RecordUpdateListener):
         priority: priority of the service
         properties: dictionary of properties (or a string holding the
                     bytes for the text field)
-        server: fully qualified name for service host (defaults to name)"""
+        server: fully qualified name for service host (defaults to name)
+        host_ttl: ttl used for A/SRV records
+        other_ttl: ttl used for PTR/TXT records"""
 
-        if not type_.endswith(service_type_name(name)):
+        if not type_.endswith(service_type_name(name, allow_underscores=True)):
             raise BadTypeInNameException
         self.type = type_
         self.name = name
@@ -1420,9 +1430,8 @@ class ServiceInfo(RecordUpdateListener):
             self.server = name
         self._properties = {}  # type: ServicePropertiesType
         self._set_properties(properties)
-        # FIXME: this is here only so that mypy doesn't complain when we set and then use the attribute when
-        # registering services. See if setting this to None by default is the right way to go.
-        self.ttl = None  # type: Optional[int]
+        self.host_ttl = host_ttl
+        self.other_ttl = other_ttl
 
     @property
     def properties(self) -> ServicePropertiesType:
@@ -1461,7 +1470,7 @@ class ServiceInfo(RecordUpdateListener):
     def _set_text(self, text):
         """Sets properties and text given a text field"""
         self.text = text
-        result = {}
+        result = {}  # type: ServicePropertiesType
         end = len(text)
         index = 0
         strs = []
@@ -1602,12 +1611,12 @@ class ServiceInfo(RecordUpdateListener):
         )
 
 
-class ZeroconfServiceTypes:
+class ZeroconfServiceTypes(ServiceListener):
     """
     Return all of the advertised services on any local networks
     """
     def __init__(self):
-        self.found_services = set()
+        self.found_services = set()  # type: Set[str]
 
     def add_service(self, zc, type_, name):
         self.found_services.add(name)
@@ -1668,7 +1677,7 @@ def new_socket(port: int = _MDNS_PORT) -> socket.socket:
     # SO_REUSEADDR should be equivalent to SO_REUSEPORT for
     # multicast UDP sockets (p 731, "TCP/IP Illustrated,
     # Volume 2"), but some BSD-derived systems require
-    # SO_REUSEPORT to be specified explicity.  Also, not all
+    # SO_REUSEPORT to be specified explicitly.  Also, not all
     # versions of Python have SO_REUSEPORT available.
     # Catch OSError and socket.error for kernel versions <3.9 because lacking
     # SO_REUSEPORT support.
@@ -1698,7 +1707,7 @@ def new_socket(port: int = _MDNS_PORT) -> socket.socket:
 
 def get_errno(e: Exception) -> int:
     assert isinstance(e, socket.error)
-    return e.args[0]
+    return cast(int, e.args[0])
 
 
 class Zeroconf(QuietLogger):
@@ -1767,7 +1776,7 @@ class Zeroconf(QuietLogger):
             self._respond_sockets.append(respond_socket)
 
         self.listeners = []  # type: List[RecordUpdateListener]
-        self.browsers = {}  # type: Dict[RecordUpdateListener, ServiceBrowser]
+        self.browsers = {}  # type: Dict[ServiceListener, ServiceBrowser]
         self.services = {}  # type: Dict[str, ServiceInfo]
         self.servicetypes = {}  # type: Dict[str, int]
 
@@ -1810,14 +1819,14 @@ class Zeroconf(QuietLogger):
             return info
         return None
 
-    def add_service_listener(self, type_: str, listener: RecordUpdateListener) -> None:
+    def add_service_listener(self, type_: str, listener: ServiceListener) -> None:
         """Adds a listener for a particular service type.  This object
-        will then have its update_record method called when information
-        arrives for that type."""
+        will then have its add_service and remove_service methods called when
+        services of that type become available and unavailable."""
         self.remove_service_listener(listener)
         self.browsers[listener] = ServiceBrowser(self, type_, listener)
 
-    def remove_service_listener(self, listener: RecordUpdateListener) -> None:
+    def remove_service_listener(self, listener: ServiceListener) -> None:
         """Removes a listener from the set that is currently listening."""
         if listener in self.browsers:
             self.browsers[listener].cancel()
@@ -1829,13 +1838,17 @@ class Zeroconf(QuietLogger):
             self.remove_service_listener(listener)
 
     def register_service(
-        self, info: ServiceInfo, ttl: int = _DNS_TTL, allow_name_change: bool = False,
+        self, info: ServiceInfo, ttl: Optional[int] = None, allow_name_change: bool = False,
     ) -> None:
-        """Registers service information to the network with a default TTL
-        of 60 seconds.  Zeroconf will then respond to requests for
-        information for that service.  The name of the service may be
-        changed if needed to make it unique on the network."""
-        info.ttl = ttl
+        """Registers service information to the network with a default TTL.
+        Zeroconf will then respond to requests for information for that
+        service.  The name of the service may be changed if needed to make
+        it unique on the network."""
+        if ttl is not None:
+            # ttl argument is used to maintain backward compatibility
+            # Setting TTLs via ServiceInfo is preferred
+            info.host_ttl = ttl
+            info.other_ttl = ttl
         self.check_service(info, allow_name_change)
         self.services[info.name.lower()] = info
         if info.type in self.servicetypes:
@@ -1852,18 +1865,18 @@ class Zeroconf(QuietLogger):
                 continue
             out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)
             out.add_answer_at_time(
-                DNSPointer(info.type, _TYPE_PTR, _CLASS_IN, ttl, info.name), 0)
+                DNSPointer(info.type, _TYPE_PTR, _CLASS_IN, info.other_ttl, info.name), 0)
             out.add_answer_at_time(
                 DNSService(info.name, _TYPE_SRV, _CLASS_IN,
-                           ttl, info.priority, info.weight, info.port,
+                           info.host_ttl, info.priority, info.weight, info.port,
                            info.server), 0)
 
             out.add_answer_at_time(
-                DNSText(info.name, _TYPE_TXT, _CLASS_IN, ttl, info.text), 0)
+                DNSText(info.name, _TYPE_TXT, _CLASS_IN, info.other_ttl, info.text), 0)
             if info.address:
                 out.add_answer_at_time(
                     DNSAddress(info.server, _TYPE_A, _CLASS_IN,
-                               ttl, info.address), 0)
+                               info.host_ttl, info.address), 0)
             self.send(out)
             i += 1
             next_time += _REGISTER_TIME
@@ -1971,7 +1984,7 @@ class Zeroconf(QuietLogger):
             self.debug = out
             out.add_question(DNSQuestion(info.type, _TYPE_PTR, _CLASS_IN))
             out.add_authorative_answer(DNSPointer(
-                info.type, _TYPE_PTR, _CLASS_IN, info.ttl, info.name))
+                info.type, _TYPE_PTR, _CLASS_IN, info.other_ttl, info.name))
             self.send(out)
             i += 1
             next_time += _CHECK_TIME
@@ -2042,14 +2055,14 @@ class Zeroconf(QuietLogger):
                             out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)
                         out.add_answer(msg, DNSPointer(
                             "_services._dns-sd._udp.local.", _TYPE_PTR,
-                            _CLASS_IN, _DNS_TTL, stype))
+                            _CLASS_IN, _DNS_OTHER_TTL, stype))
                 for service in self.services.values():
                     if question.name == service.type:
                         if out is None:
                             out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)
                         out.add_answer(msg, DNSPointer(
                             service.type, _TYPE_PTR,
-                            _CLASS_IN, service.ttl, service.name))
+                            _CLASS_IN, service.other_ttl, service.name))
             else:
                 try:
                     if out is None:
@@ -2062,7 +2075,7 @@ class Zeroconf(QuietLogger):
                                 out.add_answer(msg, DNSAddress(
                                     question.name, _TYPE_A,
                                     _CLASS_IN | _CLASS_UNIQUE,
-                                    service.ttl, service.address))
+                                    service.host_ttl, service.address))
 
                     name_to_find = question.name.lower()
                     if name_to_find not in self.services:
@@ -2072,16 +2085,16 @@ class Zeroconf(QuietLogger):
                     if question.type in (_TYPE_SRV, _TYPE_ANY):
                         out.add_answer(msg, DNSService(
                             question.name, _TYPE_SRV, _CLASS_IN | _CLASS_UNIQUE,
-                            service.ttl, service.priority, service.weight,
+                            service.host_ttl, service.priority, service.weight,
                             service.port, service.server))
                     if question.type in (_TYPE_TXT, _TYPE_ANY):
                         out.add_answer(msg, DNSText(
                             question.name, _TYPE_TXT, _CLASS_IN | _CLASS_UNIQUE,
-                            service.ttl, service.text))
+                            service.other_ttl, service.text))
                     if question.type == _TYPE_SRV:
                         out.add_additional_answer(DNSAddress(
                             service.server, _TYPE_A, _CLASS_IN | _CLASS_UNIQUE,
-                            service.ttl, service.address))
+                            service.host_ttl, service.address))
                 except Exception:  # TODO stop catching all Exceptions
                     self.log_exception_warning()
 
@@ -2115,10 +2128,10 @@ class Zeroconf(QuietLogger):
         """Ends the background threads, and prevent this instance from
         servicing further queries."""
         if not self._GLOBAL_DONE:
-            self._GLOBAL_DONE = True
             # remove service listeners
             self.remove_all_service_listeners()
             self.unregister_all_services()
+            self._GLOBAL_DONE = True
 
             # shutdown recv socket and thread
             if not self.unicast:
